#!/usr/bin/env python3
"""
Author: Fredhappyface
Date: 2020/02/19

Cat an image to the console. Inspired by img2txt
https://github.com/hit9/img2txt.git by Chao Wang
"""
from __future__ import annotations

from typing import Any, Callable, Union
import urllib.request
import os
import argparse
from pathlib import Path
from PIL import Image
LAZY_PRINT = None
try:
	from cli2gui import Cli2Gui
	from metprint import LogType, LAZY_PRINT
except ImportError:
	def Cli2Gui(*args: Any, **kwargs: Any):
		def wrapper(callingFunction: Callable[..., None]):
			def inner(*args: Any, **kwargs: Any):
				return callingFunction(*args, **kwargs)
			inner.__name__ = callingFunction.__name__
			return inner
		return wrapper

'''
if platform.system() == "Windows":
	kernel32 = ctypes.windll.kernel32
	kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

'''

THISDIR = str(Path(__file__).resolve().parent)

def openImageToPx(imageName: str, maxLen: int, hd: bool=False) -> tuple[Any, int, int]:
	"""Get an array of pixels and the dimensions of these

	Args:
		imageName (str): path of the image on the filesystem (relative of
		absolute)
		maxLen (int): maximum of width and height in chars
		hd (bool, optional): get 'hd' array of pixels. Defaults to False.

	Returns:
		int[][], int, int: 2d array of pixels, and the dimensions of the image
	"""
	if not os.path.exists(imageName):
		if LAZY_PRINT is not None:
			LAZY_PRINT(imageName + " does not exist", LogType.ERROR)
		else:
			print("- Error: " + imageName + " does not exist")
		return Image.open(THISDIR + os.sep + "broken.png").convert("RGBA").load(), 16, 16
	image = Image.open(imageName).convert("RGBA")
	initW, initH = image.size
	scale = maxLen / max(initH, initW)
	width = int(scale * initW)
	height = int(scale * initH/ (1 if hd else 2))
	image = image.resize((width, height))
	pixels = image.load()
	return pixels, width, height


def getANSIColour(rgb: tuple[int, ...]) -> int:
	"""Generate the ansi escape code based on the pixel value

	Args:
		rgb (tuple[int, ...]): int array with pixel rgb values: [r, g, b]

	Returns:
		int: ansi escape code for the colour
	"""
	websafeR = int(round((rgb[0] / 255.0) * 5))
	websafeG = int(round((rgb[1] / 255.0) * 5))
	websafeB = int(round((rgb[2] / 255.0) * 5))
	return int(((websafeR * 36) + (websafeG * 6) + websafeB) + 16)

def genANSIpx(beforeColour: Union[None, int, tuple[int, ...]], colour: Union[None, int, tuple[int, ...]], bg: bool=False,
trueColour: bool=True) -> str: # yapf: disable
	"""Generate the ansi escape string for a set of pixels with the same
	colour

	Args:
		beforeColour (Union[None, int, tuple[int, ...]]): previous colour
		colour (Union[None, int, tuple[int, ...]]): current colour
		bg (bool, optional): ansi background char. Defaults to False.
		trueColour (bool, optional): print in true colour. Defaults to True.

	Returns:
		str: string to represent char colour
	"""
	colourArr = []
	if not trueColour:
		if isinstance(beforeColour, tuple):
			beforeColour = getANSIColour(beforeColour) if beforeColour is not None else None
		if isinstance(colour, tuple):
			colour = getANSIColour(colour) if colour is not None else None

	if colour != beforeColour:
		if colour is None:
			colourArr.append("49" if bg else "39")
		elif not trueColour:
			colourArr += ["48" if bg else "38", "5", str(colour)]
		elif isinstance(colour, tuple):
			colourArr += ["48" if bg else "38", "2", str(colour[0]), str(colour[1]), str(colour[2])]
	return "\x1b[" + ";".join(colourArr) + "m" if len(colourArr) > 0 else ""


def generateHDColour(imageName: str, maxLen: int, trueColour: bool=True, char: str="\u2584") -> str:
	"""Iterate through image pixels to make a printable string

	Args:
		imageName (str): path of the image on the filesystem (relative of
		absolute)
		maxLen (int): maximum of width and height in chars
		trueColour (bool, optional): print in true colour. Defaults to True.
		char (str, optional): use this char for each pixel. Defaults to "\u2584".

	Returns:
		str: string to print
	"""
	char = "\u2584" if char in [None, ""] else char
	pixels, width, height = openImageToPx(imageName, maxLen, hd=True)
	result = ["\x1b[2K\x1b[0m"] # Clear line and reset
	beforeFgColour = None
	beforeBgColour = None
	for h in range(0, height, 2):
		for w in range(width):
			rgbaBg = pixels[w, h]
			try:
				rgbaFg = pixels[w, h+1]
			except IndexError:
				rgbaFg = pixels[w, h]
			if rgbaBg[3] != 0:
				result.append(genANSIpx(beforeBgColour, rgbaBg[:3], True, trueColour=trueColour))
				beforeBgColour = rgbaBg[:3]
			else:
				result.append(genANSIpx(beforeBgColour, None, True, trueColour=trueColour))
				beforeBgColour = None
			if rgbaFg[3] != 0:
				result.append(genANSIpx(beforeFgColour, rgbaFg[:3], trueColour=trueColour) + char)
				beforeFgColour = rgbaFg[:3]
			else:
				result.append(genANSIpx(beforeFgColour, None, trueColour=trueColour) + " ")
				beforeFgColour = None
		if h+1 != height:
			beforeFgColour = None
			beforeBgColour = None
			result.append("\x1b[39m\x1b[49m\n")

	return "".join(result)

def generateColour(imageName: str, maxLen: int, trueColour: bool=True, char: str="\u2588"):
	"""Iterate through all of the pixels in an image and construct a printable
	string

	Args:
		imageName (str): path of the image on the filesystem (relative of
		absolute)
		maxLen (int): maximum of width and height in chars
		trueColour (bool, optional): print in true colour. Defaults to True.
		char (str, optional): use this char for each pixel. Defaults to "\u2588".

	Returns:
		str: string to print
	"""
	char = "\u2588" if char in [None, ""] else char
	pixels, width, height = openImageToPx(imageName, maxLen)
	result = ["\x1b[2K\x1b[0m"] # Clear line and reset
	beforeFgColour = None
	for h in range(height):
		for w in range(width):
			rgba: tuple[int, ...] = pixels[w, h]
			if rgba[3] != 0:
				result.append(genANSIpx(beforeFgColour, rgba[:3], trueColour=trueColour) + char)
				beforeFgColour = rgba[:3]
			else:
				result.append(genANSIpx(beforeFgColour, None, trueColour=trueColour) + " ")
				beforeFgColour = None
		if h+1 != height:
			beforeFgColour = None
			result.append("\x1b[39m\n")
	return result



def generateGreyscale(imageName: str, maxLen: int):
	"""Iterate through image pixels to make a printable string

	Args:
		imageName (str): path of the image on the filesystem (relative of
		absolute)
		maxLen (int): maximum of width and height in chars

	Returns:
		str: string to print
	"""
	pixels, width, height = openImageToPx(imageName, maxLen)
	result = [""]
	color = " .;-:!>7?CO$QHNM"
	for h in range(height):
		for w in range(width):
			rgba = pixels[w, h]
			rgb: tuple[int, ...] = rgba[:3]
			result.append(color[int(sum(rgb) / 3.0 / 256.0 * 16)])
		result.append("\n")
	return "".join(result)



def handleArgs(args: argparse.Namespace):
	"""Handle arguments from the cli/ gui

	Args:
		args (argparse.Namespace): arguments
	"""
	if args.url:
		urllib.request.urlretrieve(args.image, "dowloadedImage.jpg")
		args.image = "dowloadedImage.jpg"

	maxLen = 130 if args.big else 78
	if not args.greyscale and not args.regular:
		result = generateHDColour(args.image, maxLen, not args.disable_truecolour, args.char)
	elif not args.greyscale and args.regular:
		result = generateColour(args.image, maxLen, not args.disable_truecolour, args.char)
	else:
		result = generateGreyscale(args.image, maxLen)
	print(result)

@Cli2Gui(run_function=handleArgs, image=THISDIR + os.sep + "program_icon.png", program_name="CatImage")
def cli(): # pragma: no cover
	""" CLI entry point """
	parser = argparse.ArgumentParser(description="cat an image to the terminal")
	parser.add_argument("image", type=str,
		help="image file or url")
	parser.add_argument("-u", "--url", action="store_true",
		help="image is a URL")
	parser.add_argument("-b", "--big", action="store_true",
		help="big image")
	parser.add_argument("-c", "--char", action="store",
		help="char to use in colour print use $'chr' for escaped chars")
	parser.add_argument("-t", "--disable-truecolour", action="store_true",
		help="disable output in truecolour")

	group = parser.add_argument_group("choose one of the following",
		"use the following arguments to change the look of the image")
	mxg = group.add_mutually_exclusive_group()
	mxg.add_argument("-g", "--greyscale", action="store_true",
		help="output image in greyscale (best for terminals that cannot handle ANSI)")
	mxg.add_argument("-r", "--regular", action="store_true",
		help="output image in regular definition")

	args = parser.parse_args()

	handleArgs(args)


if __name__ == "__main__": # pragma: no cover
	cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['catimage']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.2.0,<8.0.0']

extras_require = \
{'full': ['metprint>=2020.7.0,<2021.0.0', 'cli2gui>=2020.8.1,<2021.0.0']}

entry_points = \
{'console_scripts': ['catimage = catimage.catimage:cli']}

setup_kwargs = {
    'name': 'catimage',
    'version': '2020.6.0',
    'description': 'Use to cat an image to the terminal',
    'long_description': '[![Github top language](https://img.shields.io/github/languages/top/FHPythonUtils/CatImage.svg?style=for-the-badge)](../../)\n[![Codacy grade](https://img.shields.io/codacy/grade/6b3bd2d2569f44dca74a391e905042ac.svg?style=for-the-badge)](https://www.codacy.com/gh/FHPythonUtils/CatImage)\n[![Repository size](https://img.shields.io/github/repo-size/FHPythonUtils/CatImage.svg?style=for-the-badge)](../../)\n[![Issues](https://img.shields.io/github/issues/FHPythonUtils/CatImage.svg?style=for-the-badge)](../../issues)\n[![License](https://img.shields.io/github/license/FHPythonUtils/CatImage.svg?style=for-the-badge)](/LICENSE.md)\n[![Commit activity](https://img.shields.io/github/commit-activity/m/FHPythonUtils/CatImage.svg?style=for-the-badge)](../../commits/master)\n[![Last commit](https://img.shields.io/github/last-commit/FHPythonUtils/CatImage.svg?style=for-the-badge)](../../commits/master)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/catimage.svg?style=for-the-badge)](https://pypi.org/project/catimage/)\n[![PyPI Version](https://img.shields.io/pypi/v/catimage.svg?style=for-the-badge)](https://pypi.org/project/catimage/)\n\n<!-- omit in toc -->\n# CatImage\n\n<img src="readme-assets/icons/name.png" alt="Project Icon" width="750">\n\nCat an image to the terminal.\n\n- [Install Single Script](#install-single-script)\n\t- [Wget](#wget)\n\t- [Curl](#curl)\n- [Install With PIP](#install-with-pip)\n- [Example](#example)\n- [How To Use](#how-to-use)\n- [Language information](#language-information)\n\t- [Built for](#built-for)\n- [Install Python on Windows](#install-python-on-windows)\n\t- [Chocolatey](#chocolatey)\n\t- [Download](#download)\n- [Install Python on Linux](#install-python-on-linux)\n\t- [Apt](#apt)\n- [How to run](#how-to-run)\n\t- [With VSCode](#with-vscode)\n\t- [From the Terminal](#from-the-terminal)\n- [How to update, build and publish](#how-to-update-build-and-publish)\n- [Download](#download-1)\n\t- [Clone](#clone)\n\t\t- [Using The Command Line](#using-the-command-line)\n\t\t- [Using GitHub Desktop](#using-github-desktop)\n\t- [Download Zip File](#download-zip-file)\n- [Community Files](#community-files)\n\t- [License](#license)\n\t- [Changelog](#changelog)\n\t- [Code of Conduct](#code-of-conduct)\n\t- [Contributing](#contributing)\n\t- [Security](#security)\n- [Screenshots](#screenshots)\n\t- [Desktop](#desktop)\n\n## Install Single Script\n### Wget\n```bash\nwget -O /usr/bin/catimage https://raw.githubusercontent.com/FHPythonUtils/CatImage/master/catimage.py && sudo chmod 774 /usr/bin/catimage\n```\n### Curl\n```bash\ncurl -o /usr/bin/catimage https://raw.githubusercontent.com/FHPythonUtils/CatImage/master/catimage.py && sudo chmod 774 /usr/bin/catimage\n```\n\n## Install With PIP\n\n```python\npip install catimage\n```\n\nHead to https://pypi.org/project/catimage/ for more info\n\n\n## Example\nOriginal Image:\n\n<img src="readme-assets/screenshots/desktop/example-0.png" alt="Screenshot 1" width="600">\n\nGreyscale Image:\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png -g\n```\n\n<img src="readme-assets/screenshots/desktop/example-1.png" alt="Screenshot 2" width="600">\n\nRegular Definition Image:\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png -r -t\n```\n\n<img src="readme-assets/screenshots/desktop/example-2.png" alt="Screenshot 3" width="600">\n\nRegular Definition Image with \'@\':\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png -r -c @ -t\n```\n\n<img src="readme-assets/screenshots/desktop/example-3.png" alt="Screenshot 4" width="600">\n\nHD Image:\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png -t\n```\n\n<img src="readme-assets/screenshots/desktop/example-4.png" alt="Screenshot 5" width="600">\n\nHD Image with right half block:\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png -c $\'\\u2590\' -t\n```\n\n<img src="readme-assets/screenshots/desktop/example-5.png" alt="Screenshot 6" width="600">\n\nHD Image True Colour:\n```python\n./catimage.py readme-assets/screenshots/desktop/example-0.png\n```\n\n<img src="readme-assets/screenshots/desktop/example-6.png" alt="Screenshot 7" width="600">\n\n## How To Use\nUse to cat an image to the terminal, see the help text below for more\ninformation on using this tool from the command line:\n```bash\nusage: catimage [-h] [-u] [-b] [-c CHAR] [-t] [-g | -r] image\n\ncat an image to the terminal\n\npositional arguments:\n  image                 image file or url\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -u, --url             image is a URL\n  -b, --big             big image\n  -c CHAR, --char CHAR  char to use in colour print use $\'chr\' for escaped chars\n  -t, --disable-truecolour\n                        disable output in truecolour\n\nchoose one of the following:\n  use the following arguments to change the look of the image\n\n  -g, --greyscale       output image in greyscale (best for terminals that cannot handle ANSI)\n  -r, --regular         output image in regular definition\n```\n\nGUI\n\nUse the --cli2gui flag to launch a GUI\n\n<div>\n<img src="readme-assets/screenshots/desktop/gui-0.png" alt="GUI 1" width="600">\n<img src="readme-assets/screenshots/desktop/gui-1.png" alt="GUI 2" width="600">\n</div>\n\nAlternatively, import into your project and use:\n\n```python\ndef generateHDColour(imageName, maxLen, trueColour=True, char="\\u2584"):\n\t"""Iterate through image pixels to make a printable string\n\n\tArgs:\n\t\timageName (str): path of the image on the filesystem (relative of\n\t\tabsolute)\n\t\tmaxLen (int): maximum of width and height in chars\n\t\ttrueColour (bool, optional): print in true colour. Defaults to True.\n\t\tchar (str, optional): use this char for each pixel. Defaults to "\\u2584".\n\n\tReturns:\n\t\tstr: string to print\n\t"""\n\ndef generateColour(imageName, maxLen, trueColour=True, char="\\u2588"):\n\t"""Iterate through all of the pixels in an image and construct a printable\n\tstring\n\n\tArgs:\n\t\timageName (str): path of the image on the filesystem (relative of\n\t\tabsolute)\n\t\tmaxLen (int): maximum of width and height in chars\n\t\ttrueColour (bool, optional): print in true colour. Defaults to True.\n\t\tchar (str, optional): use this char for each pixel. Defaults to "\\u2588".\n\n\tReturns:\n\t\tstr: string to print\n\t"""\n\ndef generateGreyscale(imageName, maxLen):\n\t"""Iterate through image pixels to make a printable string\n\n\tArgs:\n\t\timageName (str): path of the image on the filesystem (relative of\n\t\tabsolute)\n\t\tmaxLen (int): maximum of width and height in chars\n\n\tReturns:\n\t\tstr: string to print\n\t"""\n```\n\n## Language information\n### Built for\nThis program has been written for Python 3 and has been tested with\nPython version 3.8.0 <https://www.python.org/downloads/release/python-380/>.\n\n## Install Python on Windows\n### Chocolatey\n```powershell\nchoco install python\n```\n### Download\nTo install Python, go to <https://www.python.org/> and download the latest\nversion.\n\n## Install Python on Linux\n### Apt\n```bash\nsudo apt install python3.8\n```\n\n## How to run\n### With VSCode\n1. Open the .py file in vscode\n2. Ensure a python 3.8 interpreter is selected (Ctrl+Shift+P > Python:Select Interpreter > Python 3.8)\n3. Run by pressing Ctrl+F5 (if you are prompted to install any modules, accept)\n### From the Terminal\n```bash\n./[file].py\n```\n\n## How to update, build and publish\n\n1. Ensure you have installed the following dependencies\n\tLinux\n\t```bash\n\twget dephell.org/install | python3.8\n\twget https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python3.8\n\t```\n\tWindows\n\t```powershell\n\t(wget dephell.org/install -UseBasicParsing).Content | python\n\t(wget https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py -UseBasicParsing).Content | python\n\t```\n2. Use poetry for the heavy lifting and dephell to generate requirements\n\t```bash\n\tpoetry update\n\tdephell deps convert\n\t```\n3. Build/ Publish\n\t```bash\n\tpoetry build\n\tpoetry publish\n\t```\n\tor\n\t```bash\n\tpoetry publish --build\n\t```\n\n\n## Download\n### Clone\n#### Using The Command Line\n1. Press the Clone or download button in the top right\n2. Copy the URL (link)\n3. Open the command line and change directory to where you wish to\nclone to\n4. Type \'git clone\' followed by URL in step 2\n```bash\n$ git clone https://github.com/FHPythonUtils/CatImage\n```\n\nMore information can be found at\n<https://help.github.com/en/articles/cloning-a-repository>\n\n#### Using GitHub Desktop\n1. Press the Clone or download button in the top right\n2. Click open in desktop\n3. Choose the path for where you want and click Clone\n\nMore information can be found at\n<https://help.github.com/en/desktop/contributing-to-projects/cloning-a-repository-from-github-to-github-desktop>\n\n### Download Zip File\n\n1. Download this GitHub repository\n2. Extract the zip archive\n3. Copy/ move to the desired location\n\n\n## Community Files\n### License\nMIT License\nCopyright (c) FredHappyface\n(See the [License](/LICENSE.md) for more information.)\n\n### Changelog\nSee the [Changelog](/CHANGELOG.md) for more information.\n\n### Code of Conduct\nIn the interest of fostering an open and welcoming environment, we\nas contributors and maintainers pledge to make participation in our\nproject and our community a harassment-free experience for everyone.\nPlease see the\n[Code of Conduct](https://github.com/FHPythonUtils/.github/blob/master/CODE_OF_CONDUCT.md) for more information.\n\n### Contributing\nContributions are welcome, please see the [Contributing Guidelines](https://github.com/FHPythonUtils/.github/blob/master/CONTRIBUTING.md) for more information.\n\n### Security\nThank you for improving the security of the project, please see the [Security Policy](https://github.com/FHPythonUtils/.github/blob/master/SECURITY.md) for more information.\n\n\n## Screenshots\n\n### Desktop\n<div>\n<img src="readme-assets/screenshots/desktop/screenshot-0.png" alt="Screenshot 1" width="600">\n<img src="readme-assets/screenshots/desktop/screenshot-1.png" alt="Screenshot 2" width="600">\n<img src="readme-assets/screenshots/desktop/screenshot-2.png" alt="Screenshot 3" width="600">\n</div>\n',
    'author': 'FredHappyface',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FredHappyface/Python.CatImage',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
